/* mem-debug.c */

/* code d'aide au debug du GC. */
/* pour l'instant : la fonction heap_check */

#ifdef DEBUG


#include "mlvalues.h"
#include "memory.h"
#include "roots.h"
#include "mem_misc.h"
#include "free_list.h"
#include "alloc.h"

#define HEAP_UNKNOWN 0
#define HEAP_TOSPACE 10
#define HEAP_FREELIST 11
#define HEAP_FREE 12
#define HEAP_HEADER 20
#define HEAP_VALUE 30
#define HEAP_BITS 31

#define H_BYTE(p, n) (BYTE (OBJECT (p), n))
#define H_SIZE1(p) (H_SIZE (p) + 1)
#define H_IS_WHITE(p) (IS_WHITE (OBJECT (p)))
#define H_IS_FORWARDED(p) (IS_FORWARDED (OBJECT (p)))
#define H_FORWARD_POINTER(p) (FORWARD_POINTER (OBJECT (p)))

extern address next_block;
extern address *scan_young, *copy_young;
extern address *copied_young;
extern int minor_gc_is_running;

unsigned long heap_length, young_length;
static char *major_table, *minor_table;
static address p, q;
static long i;
static long s;
int before_major_gc;


void check_error (errcode)
     int errcode;
{
  Zassert (0);
}

void check_block (p, s, errcode)
     address p;
     unsigned long s;
     int errcode;
{
  switch (H_TAG (p)){
  case TAG_STRING:
    if (s >= 3){		/* Si la chaine n'occupe qu'un mot,
				   elle peut etre ecrasee par un
				   forwarding pointer. */
      unsigned i;
      i = 4 * (s - 1);
      i = i - H_BYTE(p, i - 1);
      if (i >= 4 && H_BYTE (p, i) != 0)
        check_error (errcode + 4000);
    }
    break;
  case TAG_FLOAT:
    if (s != 1 + sizeof(double)/sizeof(value)) check_error (errcode + 1000);
    break;
  case TAG_BITVECT:
    break;
  default:
    break;
  }
}

long major_check_block (p, errcode)
     address p;
     int errcode;
{
  s = H_SIZE1 (p);
  if (s < 1) check_error (errcode + 100);
  if (s > MAX_SIZE) check_error (errcode + 200);
  if (p + s > heap_end || p + s <= heap_start) check_error (errcode + 300);
  check_block (p, s, errcode + 400);
  if (!H_IS_WHITE (p)) check_error (errcode + 500);
  return s;
}

long minor_check_block (p, errcode)
     address p;
     int errcode;
{
  s = H_SIZE1 (p);
  if (s < 2) check_error (errcode + 100);
  if (s > MAX_YOUNG_SIZE) check_error (errcode + 200);
  if (p + s > yp || p + s <= young_start) check_error (errcode + 300);
  check_block (p, s, errcode + 400);
  if (H_IS_FORWARDED (p)){
    /* verifier le forwarding pointer 
       Ca ne marche pas si on a plusieurs GC majeurs par mineur, si
       le bloc a ete forwarde, devient mort, et on deplace ensuite la copie.
    if (!IS_BLOCK (H_FORWARD_POINTER (p))) check_error (errcode + 600);
    q = ADDRESS (H_FORWARD_POINTER (p));
    if (s != H_SIZE1 (q)) check_error (errcode + 700);
    if ((H_HEADER (p) || ~BLACK) != (H_HEADER (q) || ~BLACK))
      check_error (errcode + 800);
    */
  }else{
    if (!H_IS_WHITE (p)) check_error (errcode + 500);
  }
  return s;
}

void check_value (p, errcode)
     address p;
     int errcode;
{
  if (p >= heap_start && p < heap_end){
    if (major_table [p - heap_start] != HEAP_HEADER) check_error (errcode + 100);
    if (H_SIZE1 (p) < 2) check_error (errcode + 200);
  }else if (p >= young_start && p < young_end){
    if (p >= yp) check_error (errcode + 500);
    /* cons ne remet pas a jour yp (en fonction de d3) avant d'appeler
       le gc mineur. cons a ete bricole pour le faire.
       mais le handler de signaux n'a aucune idee de la valeur de yp
       (c'est pourquoi il appelle le gc.)
    if (p >= yp) check_error (errcode + 500);
    */
    if (minor_table [p - young_start] != HEAP_HEADER) check_error (errcode + 300);
    if (H_SIZE1 (p) < 2) check_error (errcode + 400);
  }
}

void check_major_value (p, ad, errcode)
     address p, ad;
     int errcode;
{
  address *q;
  check_value (p, errcode);
  if (p >= young_start && p < young_end){
    for (q = ref_table; q < ref_index; q++){
      if (*q == ad) goto found;
    }
    for (q = scan_young; q < copy_young; q++){
      if (ad > *q && ad < *q + H_SIZE1 (*q)) goto found;
    }
    check_error (errcode + 1000);
  }
  found : ;
}

void check_minor_value (p, errcode)
     address p;
     int errcode;
{
/*  if (before_first_major_gc) check_value (p, errcode); */
}

void heap_check (is_before_major_gc)
     int is_before_major_gc;
{
  value * sp;

  /* 0. positionner les drapeaux */
  before_major_gc = is_before_major_gc;

  /* 1. les tests de base. */

  if (young_end < young_start) check_error (10);
  if (heap_end < heap_start) check_error (11);
  if (young_end > heap_start && young_start <= heap_end) check_error (12);

  if (from_end < from_start) check_error (13);
  if (to_end < to_start) check_error (14);
  if (from_start < heap_start || from_end > heap_end) check_error (15);
  if (to_start < heap_start || to_end > heap_end) check_error (16);

  if (extern_asp < arg_stack_low || extern_asp > arg_stack_high)
    check_error (17);
  if (extern_rsp < ret_stack_low || extern_rsp > ret_stack_high)
    check_error (17);
  if (yp > young_end || yp < young_start) check_error (18);

  if (scan_young < copied_young
      || scan_young > copied_young + (young_end - young_start) / 2
      || scan_young > copy_young
      || copy_young < copied_young
      || copy_young > copied_young + (young_end - young_start) / 2)
    check_error (19);

  /* 2. determiner la nature de chaque mot du tas majeur */

  heap_length = heap_end - heap_start;
  major_table = temp_alloc (heap_length);
  if (major_table == 0) return;

  for (i = 0; i < heap_length; i++){
    major_table [i] = HEAP_UNKNOWN;
  }

  for (i = to_start - heap_start; i < to_end - heap_start; i++){
    major_table [i] = HEAP_TOSPACE;
  }

  if (next_block != NULL){
    p = next_block;
    do{
      s = major_check_block (p, 20);
      if (s < 2) check_error (21);
      for (i = p - heap_start; i < p - heap_start + s; i++){
	major_table [i] = HEAP_FREELIST;
      }
      p = (address) LINK (p);
    }while (p != next_block && p != NULL);
  }

  p = q = heap_start;
  while (p < heap_end){
    if (major_table [p - heap_start] == HEAP_TOSPACE){
      if (p != to_start) check_error (22);
      p = to_end;
    }else if (major_table [p - heap_start] == HEAP_FREELIST){
      s = major_check_block (p, 23);
      p += s;
    }else{
      int contents;
      s = major_check_block (p, 24);
      major_table [p - heap_start] = HEAP_HEADER;
      contents = (H_TAG (p) >= NO_SCAN_TAG) ? HEAP_BITS : HEAP_VALUE;
      for (i = 1; i < s; i++){
	major_table [p - heap_start + i] = contents;
      }
      p += s;
    }
    if (p == q) check_error (25);
    q = p;
  }


  /* 3. idem pour le tas mineur */

  young_length = young_end - young_start;
  minor_table = temp_alloc (young_length);
  if (minor_table == 0) return;
  
  for (i = 0; i < young_length; i++){
    minor_table [i] = HEAP_UNKNOWN;
  }

  p = q = young_start;
  while (p < yp){
    int contents;
    s = minor_check_block (p, 30);
    minor_table [p - young_start] = HEAP_HEADER;
    contents = (H_TAG (p) >= NO_SCAN_TAG) ? HEAP_BITS : HEAP_VALUE;
    for (i = 1; i < s; i++){
      minor_table [p - young_start + i] = contents;
    }
    p += s;
    if (p == q) check_error (31);
    q = p;
  }
  if (p != yp) check_error (32);
  while (p < young_end){
    minor_table [p - young_start] = HEAP_FREE;
    p++;
  }


  /* 4. verifier qu'il n'y a pas de mot perdu dans un des tas. */

  for (i = 0; i < heap_length; i++){
    if (major_table [i] == HEAP_UNKNOWN) check_error (40);
  }

  for (i = 0; i < young_length; i++){
    if (minor_table [i] == HEAP_UNKNOWN) check_error (41);
  }

  /* 5. verifier la pile, les variables globales, et les racines C */

  /* La pile d'arguments */

  for (sp = extern_asp; sp < arg_stack_high; sp++) {
    if (*sp != MARK && IS_BLOCK(*sp)) {
      q = ADDRESS (*sp);
      check_value (q, 50);
    }
  }

  for (sp = extern_rsp; sp < ret_stack_high; ) {
#define retsp ((struct return_frame *) sp)
    if (!IS_BLOCK(retsp->env)) check_error(51);
    q = ADDRESS(retsp->env);
    check_value(q, 52);
    i = retsp->cache_size;
    sp += sizeof(struct return_frame) / sizeof(value);
    if (sp + i > ret_stack_high) check_error(54);
    while (i > 0) {
      if (IS_BLOCK(*sp)) {
        q = ADDRESS(*sp);
        check_value(q, 53);
      }
      sp++;
      i--;
    }
  }
  if (sp != ret_stack_high) check_error(55);
#undef retsp

  check_value(&global_data, 56);

  {
    value *C_roots_block;

    for (C_roots_block = C_roots_head; C_roots_block != NULL;
	 C_roots_block = (value *) C_roots_block [1]){
      for (sp = C_roots_block - (long) C_roots_block [0];
	   sp < C_roots_block; sp++){
	if (IS_BLOCK (*sp)){
	  q = ADDRESS (*sp);
	  check_value (q, 57);
	}
      }
    }
  }


  /* 6. verifier les pointeurs des tas */

  for (i = 0; i < heap_length; i++){
    if (major_table [i] == HEAP_VALUE){
      p = heap_start + i;
      if (IS_BLOCK (*p)){
	q = ADDRESS (*p);
	check_major_value (q, p, 60);
      }
    }
  }

  for (i = 0; i < young_length; i++){
    if (minor_table [i] == HEAP_VALUE){
      p = young_start + i;
      if (IS_BLOCK (*p)){
	q = ADDRESS (*p);
	check_minor_value (q, 64);
      }
    }
  }

  /* 7. verifier les tables des pointeurs inverses */

  {
    address *p;
    for (p = ref_table; p < ref_index; p++){
      if (*p >= heap_end || *p < heap_start) check_error (70);
      if (major_table [*p - heap_start] != HEAP_VALUE) check_error (71);
    }

  }

  temp_free ();
}

static unsigned long gc_random = 0x12345678;

unsigned long gc_debug_random ()
{
  return gc_random += 311954921; /* nombre premier */
}

#endif /* DEBUG */
