#open "misc";;
#open "constants";;
#open "hashtable";;
#open "errors";;
#open "predef";;

(* symtable.ml : numerotation des variables globales et tutti quanti *)

(* Tables hachees de numerotation *)

type 'a numtable =
  { mutable num_cnt: int;                   (* Le compteur courant *)
    mutable num_tbl: ('a * int) hashtable } (* La table *)
;;

let new_numtable size =
  { num_cnt=0; num_tbl=new_hashtable size }
;;

let find_in_numtable nt =
  find_in_assoctable nt.num_tbl

and enter_in_numtable nt key =
  let c = nt.num_cnt in
    nt.num_cnt <- succ nt.num_cnt;
    add_to_assoctable nt.num_tbl (key,c);
    c
;;

(* Les globaux *)

let global_table =
  ref (new_numtable 1 : qualified_ident numtable)
and literal_table =
  ref ([] : (int * struct_constant) list)
;;

let get_slot_for_variable qualid =
  try
    find_in_numtable !global_table qualid
  with failure "find" ->
    prerr_begline ">> ";
    prerr_qualid qualid;
    prerr_endline " referenced before being defined.";
    raise Toplevel
and get_slot_for_defined_variable qualid =
  enter_in_numtable !global_table qualid
and get_slot_for_literal cst =
  let c = (!global_table).num_cnt in
    (!global_table).num_cnt <- succ (!global_table).num_cnt;
    literal_table := (c, cst) :: !literal_table;
    c
and number_of_globals () =
  (!global_table).num_cnt
;;

(* Le type des exceptions (somme ouverte) *)

let exn_tag_table = ref(new_numtable 1 : qualified_ident numtable)
and tag_exn_table = ref( [| |] : qualified_ident vect )
and unknown_exn_name = {qual="?"; id="?"}
;;

let get_num_of_tag = function
    ConstrRegular(t,_) -> t
  | ConstrExtensible name ->
      try
        find_in_numtable !exn_tag_table name
      with failure "find" ->
        let c = enter_in_numtable !exn_tag_table name in
          if c >= vect_length !tag_exn_table then begin
            let new_tag_exn_table =
              make_vect (vect_length !tag_exn_table) unknown_exn_name
            in
              blit_vect !tag_exn_table 0 new_tag_exn_table 0
                (vect_length !tag_exn_table - 1);
              tag_exn_table := new_tag_exn_table;
              ()
          end;
          (!tag_exn_table).(c) <- name;
          c
;;

let exn_of_tag tag =
  if tag >= vect_length !tag_exn_table then failwith "exn_of_tag" else
    let name = (!tag_exn_table).(tag) in
      if name == unknown_exn_name then failwith "exn_of_tag" else name
;;

(* Les primitives C *)

let c_primitives_file = ref ""
;;
let c_primitives_table = (new_numtable 29 : string numtable)
;;
let reset_c_primitives () =
  if string_length !c_primitives_file == 0 then
    do_list (enter_in_numtable c_primitives_table) prim_c__c_primitives_list
;;
let output_c_primitives () =
  if string_length !c_primitives_file > 0 then begin
    let n = c_primitives_table.num_cnt in
    let v = make_vect n "" in
    do_hashtable (fun (s,i) -> v.(i) <- s) c_primitives_table.num_tbl;
    let oc = open_out !c_primitives_file in
    for i = 0 to n-1 do
      output_string oc v.(i);
      output_char oc `\n`
    done;
    close_out oc
  end
;;

(* Tout initialiser *)

let reset_linker_tables () =
  global_table := new_numtable 263;
  literal_table := [];
  do_list get_slot_for_defined_variable predef_variables;
  exn_tag_table := new_numtable 31;
  tag_exn_table := make_vect 50 unknown_exn_name;
  do_list (fun q -> get_num_of_tag (ConstrExtensible q)) predef_exn
;;


(* Sauver, relire les tables du linker dans un fichier *)

let save_linker_tables outstream =
  output_value outstream !global_table;
  output_value outstream !exn_tag_table;
  output_value outstream !tag_exn_table

and load_linker_tables instream =
  global_table := input_value instream;
  exn_tag_table := input_value instream;
  tag_exn_table := input_value instream;
  ()
;;

